#include "DynamicBuffer.hpp"

#include <GL/glew.h>
#include <exception>
#include <stdexcept>
#include <cstring>

#include <iostream>

namespace megu {
    template <typename T, uint32_t Y>
    DynamicBuffer<T,Y>::DynamicBuffer(size_t size, size_t array_size)
    : _max_size(size), _size(array_size) {}

    template <typename T, uint32_t Y>
    size_t DynamicBuffer<T,Y>::size() const {
        return this->_size;
    }

    template <typename T, uint32_t Y>
    size_t DynamicBuffer<T,Y>::max_size() const {
        return this->_max_size;
    }

    template <typename T, uint32_t Y>
    size_t DynamicBuffer<T,Y>::memory_size() const {
        return this->_size * sizeof(T);
    }

    template <typename T, uint32_t Y>
    size_t DynamicBuffer<T,Y>::max_memory_size() const {
        return this->_max_size * sizeof(T);
    }

    template <typename T, uint32_t Y>
    bool DynamicBuffer<T,Y>::empty() const {
        return this->_size == 0;
    }

    template <typename T, uint32_t Y>
    bool DynamicBuffer<T,Y>::full() const {
        return this->_size == this->_max_size;
    }

    template <typename T, uint32_t Y>
    void DynamicBuffer<T,Y>::set(const std::vector<T> & vertices) {
        this->_max_size = vertices.size();
        this->_size = vertices.size();
        this->generate(vertices);
    }

    template <typename T, uint32_t Y>
    void DynamicBuffer<T,Y>::insert(const std::vector<T> & vertices, size_t pos) {
        if(this->_size + vertices.size() > this->_max_size) {
            std::vector<T> data = this->extract();
            data.insert(data.begin() + pos, vertices.begin(), vertices.end());
            this->set(data);
        }
        else {
            if(this->identifier() != 0) {
                this->bind();
                uint8_t * writer = static_cast<uint8_t *>(glMapBuffer(Y, GL_READ_WRITE));
                
                if(pos < this->_size) {
                    memmove(writer + ((pos + vertices.size()) * sizeof(T)), writer + (pos * sizeof(T)), (this->_size-pos) * sizeof(T));
                }
                memmove(writer + (pos * sizeof(T)), std::vector<T>(vertices).data(), vertices.size() * sizeof(T));
                glUnmapBuffer(Y);
                
                this->_size += vertices.size();

            }
            else {
                this->generate(vertices);
            }
        }
    }  

    template <typename T, uint32_t Y>
    void DynamicBuffer<T,Y>::replace(const std::vector<T> & vertices, size_t pos) {
        if(this->empty()) {
            throw std::range_error("Buffer is empty.");
        }
        
        if(pos + vertices.size() > this->_size) {
            throw std::overflow_error("Cannot replace datas out of buffer.");
        }

        this->bind();
        glBufferSubData(Y, pos * sizeof(T),vertices.size() * sizeof(T), std::vector<T>(vertices).data());
    }

    template <typename T, uint32_t Y>
    void DynamicBuffer<T,Y>::erase(size_t from, size_t to) {
        if(this->empty()) {
            throw std::range_error("Buffer is empty.");
        }

        if(this->_size - to - from  < 0) {
            throw std::underflow_error("Cannot erase datas out of buffer.");
        }
        
        this->bind();
    
        uint8_t * writer = static_cast<uint8_t *>(glMapBuffer(Y, GL_READ_WRITE));
        memmove(writer + (from * sizeof(T)), writer + ((to+1) * sizeof(T)), (this->_size-to) * sizeof(T));
        glUnmapBuffer(Y);

        this->_size -= (to + 1 - from);
    }

    template <typename T, uint32_t Y>
    void DynamicBuffer<T,Y>::merge(const DynamicBuffer<T,Y> & vbo, size_t pos) {
        if(this->_size + vbo._size > this->_max_size) {
            throw std::overflow_error("Cannot add datas out of buffer.");
        }
            
        if(pos < this->_size) {
            this->bind();
            uint8_t * writer = static_cast<uint8_t *>(glMapBuffer(Y,GL_READ_WRITE));
            memmove(writer + (pos + vbo._size) * sizeof(T), writer + (pos * sizeof(T)), (this->_size-pos) * sizeof(T));
            glUnmapBuffer(Y);
        }

        this->_size += vbo._size;

        glBindBuffer(GL_COPY_READ_BUFFER, vbo.identifier());
        glBindBuffer(GL_COPY_WRITE_BUFFER, this->identifier());

        glCopyBufferSubData(GL_COPY_READ_BUFFER, GL_COPY_WRITE_BUFFER, 0, pos * sizeof(T),vbo._size * sizeof(T));
    }

    template <typename T, uint32_t Y>
    void DynamicBuffer<T,Y>::resize(size_t max_size) {
        if(this->_size > max_size) {
            throw std::overflow_error("Cannot set a new max size that is greater than current size");
        }

        std::vector<T> save = this->extract();
        this->_max_size = max_size;
        this->generate(save);
    }

    template <typename T, uint32_t Y>
    void DynamicBuffer<T,Y>::push_back(const std::vector<T> & vertices) {
        this->insert(vertices, this->_size);
    }

    template <typename T, uint32_t Y>
    void DynamicBuffer<T,Y>::push_front(const std::vector<T> & vertices) {
        this->insert(vertices, 0);
    }

    template <typename T, uint32_t Y>
    void DynamicBuffer<T,Y>::pop_back(size_t offset) {
        if(offset > this->_size) {
            throw std::underflow_error("Cannot erase datas out of buffer.");
        }

        if(offset > 0) {
            this->erase(this->_size-offset+1, this->_size);
        }
    }

    template <typename T, uint32_t Y>
    void DynamicBuffer<T,Y>::pop_front(size_t offset) {
        if(offset > this->_size) {
            throw std::overflow_error("Cannot add datas out of buffer.");
        }

        if(offset > 0) {
            this->erase(0, offset-1);
        }
    }

    template <typename T, uint32_t Y>
    void DynamicBuffer<T,Y>::regenerate(size_t max_size) {
        this->_max_size = max_size;
        this->generate({});
    }

    template <typename T, uint32_t Y>
    void DynamicBuffer<T,Y>::reset(const std::vector<T> & data) {
        this->set(data);
    }

    template <typename T, uint32_t Y>
    void DynamicBuffer<T,Y>::clear() {
        this->generate({});
    }

    template <typename T, uint32_t Y>
    T DynamicBuffer<T,Y>::back() const {
        if(this->empty()) {
            throw std::underflow_error("Cannot get data from empty buffer.");
        }
        
        return this->operator[](0);
    }

    template <typename T, uint32_t Y>
    T DynamicBuffer<T,Y>::front() const {
        if(this->empty()) {
            throw std::underflow_error("Cannot get data from empty buffer.");
        }
        
        return this->operator[](this->_size-1);
    }

    template <typename T, uint32_t Y>
    T DynamicBuffer<T,Y>::at(size_t index) const {
        return this->operator[](index);
    }

    template <typename T, uint32_t Y>
    std::vector<T> DynamicBuffer<T,Y>::extract() const {
        return this->extract(0, this->_size);
    }

    template <typename T, uint32_t Y>
    std::vector<T> DynamicBuffer<T,Y>::extract(size_t from,size_t to) const {
        if(to > this->_size) {
            throw std::overflow_error("Cannot extract datas out of buffer.");
        }

        this->bind();
        void * writer = glMapBuffer(Y, GL_WRITE_ONLY);
        std::vector<T> data{static_cast<T *>(writer) + from, static_cast<T *>(writer) + from + (to - from)};
        glUnmapBuffer(Y);

        return data;
    }

    template <typename T, uint32_t Y>
    T DynamicBuffer<T,Y>::operator[](size_t index) const {
        return this->extract(index,index)[0];
    }

    template <typename T, uint32_t Y>
    void DynamicBuffer<T,Y>::modify(std::function<void(void *)> modif) {
        this->bind();
        modif(glMapBuffer(Y, GL_READ_WRITE));
        glUnmapBuffer(Y);
    }

    template <typename T, uint32_t Y>
    void DynamicBuffer<T,Y>::operator()(std::function<void(void *)> modif) {
        this->modify(modif);
    }

    template <typename T, uint32_t Y>
    void DynamicBuffer<T,Y>::operator<<(const DynamicBuffer<T,Y> & buffer) {
        this->merge(buffer);
    }
}