#include "VerticeBuffer.hpp"

#include <list>

namespace megu {
    VerticeBuffer::VerticeBuffer(const VertexArray & vao, const Layout_t & attributes, size_t size, EditMode mode)
    : DynamicBuffer(size), _vao(&vao), _id(0), _mode(mode), _layout(attributes) {
        this->generate({});
    } 

    VerticeBuffer::VerticeBuffer(const VertexArray & vao, const Layout_t & attributes, const std::vector<float> & vertices, EditMode mode)
    : DynamicBuffer(vertices.size(), vertices.size()), _vao(&vao), _id(0), _mode(mode), _layout(attributes) {
        this->generate(vertices);
    }

    VerticeBuffer::~VerticeBuffer() {
        this->unbind();
        glDeleteBuffers(1, &this->_id);
    }

    size_t VerticeBuffer::weight() const {
        return layout::weight(this->_layout);
    }

    size_t VerticeBuffer::vertices() const {
        return this->size() / this->weight();
    }

    void VerticeBuffer::bind() const {
        if(this->_id != 0) {
            this->_vao->bind();
            glBindBuffer(GL_ARRAY_BUFFER, this->_id);
        }
    }

    void VerticeBuffer::unbind() const {
        GLint vbo = 0;
        glGetIntegerv(GL_ARRAY_BUFFER, &vbo);
        if(static_cast<GLuint>(vbo) == this->_id) {
            this->_vao->unbind();
            glBindBuffer(GL_ARRAY_BUFFER, 0);
        }
    }

    void VerticeBuffer::generate(const std::vector<float> & vertices) {
        this->_vao->bind();

        if(this->_id != 0) {
            glBindBuffer(GL_ARRAY_BUFFER,0);
            glDeleteBuffers(1, &this->_id);
        }

        glGenBuffers(1,&this->_id);
        glBindBuffer(GL_ARRAY_BUFFER, this->_id);
        glBufferData(GL_ARRAY_BUFFER, this->max_size()*sizeof(float), vertices.size() == 0 ? NULL : vertices.data(), this->_mode);

        uint32_t layer = 0,accumulateWeight = 0,totalWeight = static_cast<uint32_t>(this->weight());
        for(uint8_t attribute : this->_layout) {
            glVertexAttribPointer(layer, attribute, GL_FLOAT, GL_FALSE, totalWeight * sizeof(float), (void*)(accumulateWeight * sizeof(float)));
            glEnableVertexAttribArray(layer++);
            accumulateWeight += attribute;
        }
    }

    bool VerticeBuffer::compare(const Layout_t & layout) const {
        return this->_layout == layout;
    }

    bool VerticeBuffer::operator==(const VerticeBuffer & vbo) const {
        return this->_id == vbo._id;
    }

    bool VerticeBuffer::operator!=(const VerticeBuffer & vbo) const {
        return this->_id != vbo._id;
    }

    bool VerticeBuffer::operator>(const VerticeBuffer & vbo) const {
        return this->_id > vbo._id;
    }

    bool VerticeBuffer::operator>=(const VerticeBuffer & vbo) const {
        return this->_id >= vbo._id;
    }

    bool VerticeBuffer::operator<(const VerticeBuffer & vbo) const {
        return this->_id < vbo._id;
    }

    bool VerticeBuffer::operator<=(const VerticeBuffer & vbo) const {
        return this->_id <= vbo._id;
    }

    bool VerticeBuffer::operator()(const VerticeBuffer & vbo) const {
        return *this < vbo;
    }

    std::strong_ordering VerticeBuffer::operator<=>(const VerticeBuffer & vbo) const {
        return this->_id <=> vbo._id; 
    }

    VerticeBuffer & operator<<(VerticeBuffer & vbo,const std::initializer_list<float> & data) {
        vbo.push_back(data);
        return vbo;
    }

    VerticeBuffer & operator<<(VerticeBuffer & vbo,const std::vector<float> & data) {
        vbo.push_back(data);
        return vbo;
    }
}