#include "Walkaround.hpp"

#include <glm/gtc/matrix_transform.hpp>

namespace megu {
    Walkaround::Walkaround()
    : Walkaround(glm::vec3(0.f, 0.f, 3.f)) {}

    Walkaround::Walkaround(const glm::vec3 & position, double fov, double ratio, double near, double far)
    : _view(1.f), 
      _projection(1.f), 
      _position(position), 
      _up(0.f, 1.f, 0.f),
      _pitch(0.f),
      _yaw(-90.f),
      _roll(0.f),
      _fov(fov),
      _ratio(ratio),
      _near(near),
      _far(far),
      _update(Update_Flags::NONE) {
        this->update(Update_Flags::VIEW | Update_Flags::PROJECTION);
    }

    void Walkaround::update(uint8_t flags) const {
        this->_update |= flags;
        if(this->_update != Update_Flags::NONE) {
            if((this->_update & VIEW) != 0) {
                glm::vec3 front;
                front.x = static_cast<float>(std::cos(glm::radians(this->_yaw)) * std::cos(glm::radians(this->_pitch)));
                front.y = static_cast<float>(std::sin(glm::radians(this->_pitch)));
                front.z = static_cast<float>(std::sin(glm::radians(this->_yaw)) * std::cos(glm::radians(this->_pitch)));
            
                glm::vec right = glm::normalize(glm::cross(front, this->_up));
                glm::mat4 roll_mat = glm::rotate(glm::mat4(1.f), (float)glm::radians(this->_roll), front);
                
                this->_up = glm::normalize(glm::cross(right, front));
                this->_up = glm::mat3(roll_mat) * this->_up;
               
                this->_view = glm::lookAt(this->_position, this->_position + front, this->_up);
            }

            if((this->_update & PROJECTION) != 0) {
                this->_projection = glm::perspective(glm::radians(this->_fov), this->_ratio, this->_near, this->_far);
            }
        }
        this->_update = Update_Flags::NONE;
    }

    void Walkaround::setPosition(const glm::vec3 & position) {
        this->_position = position;
        this->_update |= Update_Flags::VIEW;
    }

    void Walkaround::setPitch(double pitch) {
        this->_pitch = pitch;
        this->_update |= Update_Flags::VIEW;
    }

    void Walkaround::setYaw(double yaw) {
        this->_yaw = yaw;
        this->_update |= Update_Flags::VIEW;
    }

    void Walkaround::setRoll(double roll) {
        this->_roll = std::fmod(roll, 360.0);
        this->_update |= Update_Flags::VIEW;
    }

    void Walkaround::setRatio(double ratio) {
        this->_ratio = ratio;
        this->_update |= Update_Flags::PROJECTION;
    }

    void Walkaround::setDistance(double near, double far) {
        this->_near = near;
        this->_far = far;
        this->_update |= Update_Flags::PROJECTION;
    }

    void Walkaround::move(const glm::vec3 & movement) {
        this->_position += movement;
        this->_update |= Update_Flags::VIEW;
    }

    void Walkaround::rotate(const glm::vec3 & rotation) {
        this->_pitch += rotation.x;
        this->_yaw += rotation.y;
        this->_roll += rotation.z;
        this->_update |= Update_Flags::VIEW;
    }

    const glm::mat4 & Walkaround::view() const {
        this->update();
        return this->_view;
    }

    const glm::mat4 & Walkaround::projection() const {
        this->update();
        return this->_projection;
    }
}