#include "Transformable.hpp"

#include <glm/gtc/matrix_transform.hpp>

#include <iostream>

namespace megu {
    Transformable::Transformable(const glm::mat4x4 & matrix)
    : _model(matrix), 
      _position(0.f), 
      _rotation(0.f), 
      _scaling(1.f), 
      _origine(0.f), 
      _updated(true) {}

    Transformable::Transformable(const glm::vec3 & position, const glm::vec3 & origine)
    : _model(1.0f), 
      _position(position), 
      _rotation(0.f), 
      _scaling(1.f), 
      _origine(origine), 
      _updated(false) {}

    Transformable::Transformable(const Transformable & src)
    : _model(src._model), 
      _position(src._position), 
      _rotation(src._rotation), 
      _scaling(src._scaling), 
      _origine(src._origine), 
      _updated(false) {}

    Transformable::Transformable(const Transformable && src)
    : _model(std::move(src._model)), 
      _position(std::move(src._position)), 
      _rotation(std::move(src._rotation)), 
      _scaling(std::move(src._scaling)), 
      _origine(std::move(src._origine)), 
      _updated(false) {}

    Transformable Transformable::operator=(const Transformable & src) {
        this->_model    = src._model;
        this->_origine  = src._origine;
        this->_position = src._position;
        this->_rotation = src._rotation;
        this->_scaling  = src._scaling;
        this->_updated  = src._updated;
        return *this;
    }

    Transformable Transformable::operator=(const Transformable && src) {
        this->_model    = std::move(src._model);
        this->_origine  = std::move(src._origine);
        this->_position = std::move(src._position);
        this->_rotation = std::move(src._rotation);
        this->_scaling  = std::move(src._scaling);
        this->_updated  = std::move(src._updated);
        return *this;
    }

    void Transformable::setPosition(const glm::vec3 & position) {
        this->_position = position;
        this->_updated = true;
    }

    void Transformable::setX(float x) {
        this->_position.x = x;
        this->_updated = true;
    }

    void Transformable::setY(float y) {
        this->_position.y = y;
        this->_updated = true;
    }

    void Transformable::setZ(float z) {
        this->_position.z = z;
        this->_updated = true;
    }

    void Transformable::setRotation(float angle, Axis axis) {
        if((Axis::X & axis) != 0) {
            this->_rotation.x = angle; 
        }

        if((Axis::Y & axis) != 0) {
            this->_rotation.y = angle; 
        }

        if((Axis::Z & axis) != 0) {
            this->_rotation.z = angle; 
        }

        this->_updated = true;
    }

    void Transformable::setScaling(const glm::vec3 & scaling) {
        this->_scaling = scaling;
        this->_updated = true;
    }

    void Transformable::move(const glm::vec3 & position) {
        this->_position += position;
        this->_updated = true;
    }

    void Transformable::rotate(float angle, Axis axis) {
        if((Axis::X & axis) != 0) {
            this->_rotation.x += angle; 
        }

        if((Axis::Y & axis) != 0) {
            this->_rotation.y += angle; 
        }

        if((Axis::Z & axis) != 0) {
            this->_rotation.z += angle; 
        }

        this->_updated = true;
    }

    void Transformable::scale(const glm::vec3 & scaling) {
        this->_scaling += scaling;
        this->_updated = true;
    }

    void Transformable::update() const {
        glm::mat4x4 model = glm::mat4x4(1.f);

        model = glm::translate(model, this->_position);
        model = glm::translate(model, this->_origine);
           
        model = glm::rotate(model, glm::radians(this->_rotation.z), {0.f, 0.f, 1.f});
        model = glm::rotate(model, glm::radians(this->_rotation.y), {0.f, 1.f, 0.f});
        model = glm::rotate(model, glm::radians(this->_rotation.x), {1.f, 0.f, 0.f});

        model = glm::translate(model, -this->_origine);
        model = glm::scale(model, this->_scaling);
            
        this->_model = model;
        this->_updated = false;
    }

    const glm::mat4x4 & Transformable::model() const {
        if(this->_updated) {
            this->update();
        }
        return this->_model;
    }

    glm::mat4 Transformable::inverse_model() const {
        if(this->_updated) {
            this->update();
        }
        return glm::inverse(this->_model);
    }

    bool Transformable::operator==(const Transformable & t) const {
        return this->_position == t._position;
    }

    bool Transformable::operator!=(const Transformable & t) const {
        return this->_position != t._position;
    }

    bool Transformable::operator>=(const Transformable & t) const {
        return this->_position.x >= t._position.x &&  this->_position.y >= t._position.y && this->_position.z >= t._position.z;
    }

    bool Transformable::operator<=(const Transformable & t) const {
        return this->_position.x >= t._position.x &&  this->_position.y >= t._position.y && this->_position.z >= t._position.z;
    }

    bool Transformable::operator>(const Transformable & t) const {
        return this->_position.x > t._position.x &&  this->_position.y > t._position.y && this->_position.z > t._position.z;
    }

    bool Transformable::operator<(const Transformable & t) const {
        return this->_position.x < t._position.x && this->_position.y < t._position.y && this->_position.z < t._position.z;
    }
}