#pragma once

#include <glm/vec3.hpp>
#include <glm/mat4x4.hpp>

namespace megu {
    class Transformable {
        public:
            enum class Axis : uint8_t {
                X = 1,
                Y = 2, 
                Z = 4
            };

            inline friend uint8_t operator&(Axis a, Axis b) {return static_cast<uint8_t>(a) & static_cast<uint8_t>(b);}
            inline friend uint8_t operator|(Axis a, Axis b) {return static_cast<uint8_t>(a) | static_cast<uint8_t>(b);}
            inline friend uint8_t operator^(Axis a, Axis b) {return static_cast<uint8_t>(a) ^ static_cast<uint8_t>(b);}

            Transformable(const glm::mat4 & = glm::mat4(1.f));
            Transformable(const glm::vec3 &, const glm::vec3 & = glm::vec3(0.f));
            Transformable(const Transformable &);
            Transformable(const Transformable &&);
            Transformable operator=(const Transformable &);
            Transformable operator=(const Transformable &&);
           ~Transformable() = default;

            inline const glm::vec3 & position() const {return this->_position;}
            inline const glm::vec3 & rotation() const {return this->_rotation;}
            inline const glm::vec3 & scaling() const {return this->_scaling;}
            inline const glm::vec3 & origine() const {return this->_origine;}

            inline glm::vec3 & position() {return this->_position;}
            inline glm::vec3 & rotation() {return this->_rotation;}
            inline glm::vec3 & scaling() {return this->_scaling;}
            inline glm::vec3 & origine() {return this->_origine;}

            inline float x() const {return this->_position.x;}
            inline float y() const {return this->_position.y;}
            inline float z() const {return this->_position.z;}

            inline float pitch() const {return this->_rotation.x;}
            inline float yaw()   const {return this->_rotation.y;}
            inline float roll()  const {return this->_rotation.z;}

            void setPosition(const glm::vec3 &);
            void setRotation(float, Axis);
            void setScaling(const glm::vec3 &);
            void setX(float);
            void setY(float);
            void setZ(float);

            inline void setMatrix(const glm::mat4 & matrix) {this->_model = matrix;}
            inline void setOrigine(const glm::vec3 & position) {this->_origine = position;}
            inline void setPosition(float x, float y, float z = 0.f) {this->setPosition({x, y, z});}
            inline void setScaling(float x, float y, float z = 0.f) {this->setScaling({x, y, z});}
            inline void setOrigine(float x, float y, float z = 0.f) {this->setOrigine({x, y, z});}

            void move(const glm::vec3 &);
            void rotate(float, Axis);
            void scale(const glm::vec3 &);

            inline void translate(const glm::vec3 & translation) {this->move(translation);}
            inline void move(float x, float y, float z = 0.f) {this->move({x, y, z});}
            inline void scale(float x, float y, float z = 0.f) {this->scale({x, y, z});}

            bool operator==(const Transformable &) const;
            bool operator!=(const Transformable &) const;
            bool operator>=(const Transformable &) const;
            bool operator<=(const Transformable &) const;
            bool operator>(const Transformable &) const;
            bool operator<(const Transformable &) const;
            
            virtual const glm::mat4x4 & model() const;
            glm::mat4 inverse_model() const;

        protected:
            void update() const;

        private:
            mutable glm::mat4 _model;
            glm::vec3 _position;
            glm::vec3 _rotation;
            glm::vec3 _scaling;
            glm::vec3 _origine;
            mutable bool _updated;
    };
}