#include "Texture.hpp"

#include <exception>
#include <iostream>

namespace megu {
    Texture::Texture(GLuint slot, GLuint type) 
    : _id(0), _type(type), _slot(slot), _width(0), _height(0) {
        glGenTextures(1, &this->_id);
    }

    Texture::Texture(const std::filesystem::path & path, GLuint slot)
    : _id(0), _type(GL_RED), _slot(slot) {
        this->store(TextureBuffer(path));
    }

    Texture::Texture(const Texture & texture) 
    : _id(texture._id), _type(texture._type), _slot(texture._slot) {}

    Texture & Texture::operator=(const Texture & texture) {
        this->_id = texture._id;
        this->_slot = texture._slot;
        this->_type = texture._type;
        return *this;
    }

    Texture::~Texture() {
        glDeleteTextures(1, &this->_id);
    }

    void Texture::bind() const {
        glActiveTexture(GL_TEXTURE0 + this->_slot);
        glBindTexture(this->_type, this->_id);
    }

    void Texture::bind(GLuint slot) const {
        this->_slot = slot;
        this->bind();
    }

    void Texture::changeSlot(GLuint slot) {
        this->_slot = slot;
    }

    void Texture::setWraping(Wraping wraping, uint32_t flags) {
        glBindTexture(this->_type, this->_id);
        if(flags != 0) {
            if((flags & S) != 0) {
                glTexParameteri(this->_type, GL_TEXTURE_WRAP_S, wraping);
            }

            if((flags & T) != 0) {
                glTexParameteri(this->_type, GL_TEXTURE_WRAP_T, wraping);
            }

            if((flags & R) != 0) {
                glTexParameteri(this->_type, GL_TEXTURE_WRAP_R, wraping);
            }
        }
    }

    void Texture::setSmoothing(bool state) {
        glBindTexture(this->_type, this->_id);
        glTexParameteri(this->_type, GL_TEXTURE_MIN_FILTER, state ? GL_LINEAR : GL_NEAREST);
        glTexParameteri(this->_type, GL_TEXTURE_MAG_FILTER, state ? GL_LINEAR : GL_NEAREST);
    }

    void Texture::store(GLsizei width, GLsizei height, const GLubyte * data, uint32_t format) {
        if(this->valid()) {
            glDeleteTextures(1, &this->_id);
            glGenTextures(1, &this->_id);
        }

        this->_width = width;
        this->_height = height;
        
        glBindTexture(this->_type, this->_id);
        glTexImage2D(this->_type, 0, format, width, height, 0, format, GL_UNSIGNED_BYTE, data);
        glGenerateMipmap(this->_type);

        this->setWraping(CLAMP_TO_BORDER, S | T );
        this->setSmoothing(false);
    }

    void Texture::store(const TextureBuffer & texture) {
        if(texture.empty()) {
            throw std::runtime_error("Cannot store texture.");
        }

        this->store(texture.width(), texture.height(), texture.data_ptr(), texture.format());
    }

    bool Texture::valid() const {
        return this->_id != 0;
    }

    bool Texture::operator==(const Texture & texture) const {
        return this->_id == texture._id;
    }

    /*bool Texture::operator!=(const Texture & texture) const {
        return this->_id != texture._id;
    }*/

    bool Texture::operator>=(const Texture & texture) const {
        return this->_id >= texture._id;
    }

    bool Texture::operator<=(const Texture & texture) const {
        return this->_id <= texture._id;
    }

    bool Texture::operator>(const Texture & texture) const {
        return this->_id > texture._id;
    }

    bool Texture::operator<(const Texture & texture) const {
        return this->_id < texture._id;
    }

    std::strong_ordering Texture::operator<=>(const Texture & texture) const {
        return this->_id <=> texture._id;
    }

    void Texture::Unbind(GLuint type) {
        glBindTexture(type, 0);
    }

    GLint Texture::CountSlots() {
        GLint units = 0;
        glGetIntegerv(GL_MAX_TEXTURE_IMAGE_UNITS, &units);
        return units;
    }

    void Texture::SetDisabelingByteAligment(bool state) {
        glPixelStorei(GL_UNPACK_ALIGNMENT, state);
    }
}