#include "Engine.hpp"

#define NORMALIZE(X) X/255.f

namespace megu {
    GraphicEngine::GraphicEngine(Window & window, float w, float h) 
    : _window(window) {
        glViewport(0, 0, window.width(), window.height());
    } 

    GraphicEngine::GraphicEngine(Window & window) 
    : _window(window) {
        glViewport(0, 0, window.width(), window.height());
    }

    void GraphicEngine::push(Priority priority, Renderer & renderer) {
        this->_layers[priority] = std::make_unique<Layer>(renderer);
    }

    void GraphicEngine::remove(Priority priority) {
        auto it = this->_layers.find(priority);
        if(it != this->_layers.end()) {
            this->_layers.erase(it);
        }
    }

    void GraphicEngine::remove(const Identifiable & identifiable) {
        for(auto & [priority, layer] : this->_layers) {
            auto r = layer.get()->get(identifiable);
            if(r.has_value()) {
                layer.get()->remove(identifiable);
            }
        }
    }

    std::optional<std::reference_wrapper<const Layer>> GraphicEngine::get(Priority priority) const {
        const auto it = this->_layers.find(priority);
        if(it != this->_layers.end()) {
            return *it->second;
        }
        return {};
    }

    std::optional<std::reference_wrapper<const Renderable>> GraphicEngine::get(const Identifiable & id) const {
        for(const auto & [p, layer] : this->_layers) {
            auto renderable = layer->get(id);
            if(renderable.has_value()) {
                return renderable;
            }
        }
        return {};
    }

    void GraphicEngine::step() {
        if(this->_window.isOpen()) {
            // Draw Layers
            TextureArray textures;
            for(auto & [priority, layer] : this->_layers) {
                const glm::vec2 & dimension = layer->renderer().dimension();
                glViewport(0, 0, static_cast<GLsizei>(dimension.x), static_cast<GLsizei>(dimension.y));
                textures.push_back(layer->draw(this->_window, textures));
            }

            // Merge Textures
            FrameBuffer::BindDefaultFrameBuffer();
            this->_merger.render(this->_window, textures);
            this->_window.swapBuffers();
        }
    }

    void GraphicEngine::step(Priority priority) {
        if(this->_window.isOpen()) {
            TextureArray textures;

            auto & layer = this->_layers[priority];
            const glm::vec2 & dimension = layer->renderer().dimension();
            glViewport(0, 0, static_cast<GLsizei>(dimension.x), static_cast<GLsizei>(dimension.y));
            textures.push_back(layer->draw(this->_window, textures));

            FrameBuffer::BindDefaultFrameBuffer();
            this->_merger.render(this->_window, textures);
            this->_window.swapBuffers();
        }
    }

    void GraphicEngine::setClearColor(float x, float y, float z) {
        for(auto & [priority, layer] : this->_layers) {
            layer.get()->renderer().setClearColor(x, y, z);
        }
    } 
}