#include "Engine.hpp"

#include <iostream>

namespace megu {
    PhysicEngine::PhysicEngine() {}

    void PhysicEngine::push(Priority p, TangibleStatic & t) {
        this->_statics[p].push_back(t);
        if(!this->_dynamic.contains(p)) {
            this->_dynamic[p] = {};
        }
    }

    void PhysicEngine::push(Priority p, TangibleMovable & t) {
        this->_dynamic[p].push_back(t);
        if(!this->_statics.contains(p)) {
            this->_statics[p] = {};
        }
    }

    void PhysicEngine::remove(TangibleStatic & t) {
        for(auto & [p, objs] : this->_statics) {
            auto it = std::find(objs.begin(), objs.end(), t);
            if(it != objs.end()) {
                objs.erase(it);
                break;
            }
        }
    }

    void PhysicEngine::remove(TangibleMovable & t) {
        for(auto & [p, objs] : this->_dynamic) {
            auto it = std::find(objs.begin(), objs.end(), t);
            if(it != objs.end()) {
                objs.erase(it);
                break;
            }
        }
    }

    void PhysicEngine::step(double time) {
        this->_collisions.clear();
        for(const auto & [priority, layer] : this->_dynamic) {
            this->step(time, priority);
        }
    }

    void PhysicEngine::step(double delta, Priority priority) {
        for(auto & source : this->_dynamic[priority]) {
            source.get().updatePhysic(delta);
                
            for(auto & target : this->_statics[priority]) {
                if(source.get() | target) {
                    this->_collisions.insert(Collision(source, target));
                    this->_collisions.insert(Collision(target, source));
                }
            }

            for(auto & target : this->_dynamic[priority]) {
                if(source.get() | target) {
                    this->_collisions.insert(Collision(source, target));
                }
            }
        }
    }

    std::optional<SquareBox> PhysicEngine::checkCollision(const SquareBox & box, Priority priority) const {
        for(const auto & source : this->_statics.at(priority)) {
            auto obox = source.get().isColliding(box);
            if(obox.has_value()) {
                return obox;
            }
        }

        for(const auto & source : this->_dynamic.at(priority)) {
            auto obox = source.get().isColliding(box);
            if(obox.has_value()) {
                return obox;
            }
        }

        return {};
    }

    std::optional<std::reference_wrapper<const Tangible>> PhysicEngine::get(const Identifiable & id) const {
        for(auto & [priotiy, objects] : this->_statics) {
            for(auto & object : objects) {
                if(object.get().id() == id.id()) {
                    return object.get();
                }
            }
        }

        for(auto & [priotiy, objects] : this->_dynamic) {
            for(auto & object : objects) {
                if(object.get().id() == id.id()) {
                    return object.get();
                }
            }
        }

        return {};
    }
}