#include "Sprite.hpp"

namespace megu::kernel {
    GraphicLinker<megu::Sprite> Sprite::_Linker = GraphicLinker<megu::Sprite>();

    Sprite::Sprite()
    : megu::Sprite() {}

    Sprite::Sprite(const std::filesystem::path & path, const Frame & frame)
    : megu::Sprite(path, frame), _animation(0), _index(0) {
        if(!Sprite::_Linker.haveModule()) {
            Sprite::_Linker.setModule(new Sprite_Module{});
        }
    }

    const Frame_List & Sprite::getCurrentAnimations(size_t name) const {
        if(this->_frames.contains(name)) {
            return this->_frames.at(name);
        }
        throw std::runtime_error("Missing animation : " + name);
    }

    const Frame & Sprite::getCurrentFrame(size_t name, size_t index) const {
        auto & frames = this->getCurrentAnimations(name);
        if(frames.size() > index) {
            return frames.at(index);
        }
        throw std::runtime_error("Missing frame for animation : " + name);
    }

    void Sprite::setCurrentAnimation(size_t name) {
        if(this->_frames.contains(name)) {
            this->_animation = name;
            this->_previous = 0.0;
        }
        else {
            throw std::runtime_error("Cannot set inexisting animation : " + name);
        }
    }

    void Sprite::pushAnimation(size_t name, const Frame_List & frames) {
        for(auto & frame : frames) {
            this->pushAnimation(name, frame);
        }
    }

    void Sprite::pushAnimation(size_t name, const Frame & frame) {
        this->_frames[name].push_back(frame);
    }

    void Sprite::pushAnimation(size_t name, float x, float y, float w, float h) {
        this->_frames[name].push_back({x, y, w, h});
    }

    void Sprite::removeAnimation(size_t name) {
        if(this->_frames.contains(name)) {
            this->_frames.erase(name);
        }
    }

    void Sprite::removeAnimation(size_t name, size_t index) {
        if(this->_frames.contains(name) && this->_frames[name].size() > index) {
            this->_frames[name].erase(this->_frames[name].begin() + index);
        }
    }

    void Sprite::update(double time) {
        if(this->_animation != 0) {
            if(this->_previous == 0.0) {
                this->setFrame(this->_frames[this->_animation][this->_index]);
                this->_previous = time;
            }
            else {
                double current = this->_previous - time;
                if(current > this->_duration) {
                    this->_index = (this->_index + 1) % this->_frames[this->_animation].size();
                    this->setFrame(this->_frames[this->_animation][this->_index]);
                }
            }
        }
    }

    void Sprite::apply(GraphicEngine & engine) {
        Sprite::_Linker.link(*this, this->getLayerPriority(), this->getObjectPriority());
        Sprite::_Linker.link(engine);
    }

    void Sprite::unapply(GraphicEngine & engine) {
        Sprite::_Linker.unlink(*this);
        Sprite::_Linker.link(engine);
    }
}