#include "TileArray.hpp"

#include <kernel/front/Kernel.hpp>

namespace megu::kernel {
    TileArray::TileArray(float x, float y, float w, float h, CollideLambda cl, UpdateLambda ul) 
    : TangibleStatic(Position(x, y), Dimension(w, h, 0.f)), _collide(cl), _update(ul) {}

    std::optional<std::reference_wrapper<const Tile>> TileArray::at(const Position & position) const {
        for(auto & [pos, tile] : this->_tiles) {
            if(pos == position) {
                return tile;
            }
        }
        
        return std::optional<std::reference_wrapper<const Tile>>();
    }

    void TileArray::push(const Position & position, Tile & tile) {
        this->_tiles.insert({position, tile});
    }

    void TileArray::remove(const Tile & tile) {
        for(auto it = this->_tiles.begin(); it != this->_tiles.end();) {
            if(it->second.get().id() == tile.id()) {
                it = this->_tiles.erase(it);
            }
            else {
                ++it;
            }
        }
    }

    void TileArray::onCollide(Kernel & kernel, PhysicEngine & engine, Identifiable & id, Physical & physical, double time) {
        if(this->_collide == nullptr) {
            return;
        }
        
        auto & tangible = engine.get(physical);
        for(auto & [positon, tile] : this->_tiles) {
            SquareBox box = SquareBox(positon, tile.get().getBox().getDimension());
            if(tangible.isColliding(box)) {
                this->_collide(kernel, engine, id, tile, time);
            }
        }
    }

    std::optional<SquareBox> TileArray::isColliding(const SquareBox & box) const {
        for(auto & [positon, tile] : this->_tiles) {
            SquareBox tbox = SquareBox(positon, tile.get().getBox().getDimension());
            if(box.intersect(tbox)) {
                return tbox;
            }
        }
        return {};
    }

    void TileArray::updatePhysic(double time) const {
        if(this->_update != nullptr) {
            this->_update(time);
        }
    }

    void TileArray::apply(PhysicEngine & engine) {
        engine.get().push(this->getLayer(), *this);
    }

    void TileArray::unapply(PhysicEngine & engine) {
        engine.get().remove(*this);
    }

    void TileArray::setCollideLambda(TileCollideLambda lambda) {
        this->_collide = lambda;
    }

    void TileArray::setUpdateLambda(UpdateLambda lambda) {
        this->_update = lambda;
    }
}